#pragma rtGlobals=1		// Use modern global access method.

// Synatax Highlighter (aka Syntax Colorer)
// written by Richard C. Gerkin, 2010.  
//
// This procedure file uses XML reading/writing to edit XML files used by third party text editors to control syntax highlighting/coloring of keywords, and adds/modifies settings for coloring Igor procedure files (.ipf extension).  
// Usage: 
//        SetSyntaxColoring() or SetSyntaxColoring(1).  The find the XML file of interest in the dialog that appears.  
// Requirements: 
//        The XMLUtils XOP, written by Andy Nelson and available on IgorExchange (http://www.igorexchange.com).  
//        A third party text editor (e.g. Notepad++) that uses XML files to control syntax highlighting.  Currently it only supports Notepad++, but support for other text editors can be added.  
// Example usage with Notepad++:  
//        You will first need to add an entry for Igor to the XML file.  In Notepad++, this can be done by going to View / User-Defined Dialogue / Create New... and entering "igor" (without the quotes).  
//        Then quit Notepad++, run SetSyntaxColoring() in Igor, find the XML file (called userDefineLang and located in the Window's user's Application Data folder) and hit "Open".  Restart Notepad++, open 
//        an Igor procedure file and it should be beautifully colored!  

// This is the main function by which a 3rd party editor's syntax coloring XML control file is updated to include Igor syntax coloring.  
Function SetSyntaxColoring([userColors])
	variable userColors // Use the user's own color settings (1) or the default Igor color settings (0).  
	
	// Prompts the user to find the XML control file, which will usually be in a user directory somewhere.  
	variable refNum
	Open /D/R/M="Choose an XML file that controls syntax highlighting" /T=".xml" refNum
	if(!strlen(s_filename))
		return -1
	endif
	variable fileID=XMLOpenfile(S_filename)
	if(fileID<0)
		return -2
	endif
	
	XMLelemlist(fileID) // Create a text wave of XML elements.  
	wave /t w_elementlist
	string type=w_elementlist[0][3] // Assumes that the root XML element will tell us what kind of text editor we are dealing with, e.g. Notepad++.  
	string words="",colors=""
	variable err=XMLValues(words,colors,userColors) // Get a list of words to be colored and the colors to use.  
	err+=SetXMLValues(type,fileID,words,colors) // Update the XML file to include these words and colors.  
	killwaves /z w_elementlist
	XMLCloseFile(fileID,err==0) // Close the file and save it if there were no errors.  
End

static Function XMLValues(wordList,colorList,userColors)
	string &wordList,&colorList
	variable userColors
	
	// Generate key-value pairs for word types and the colors that should be used.  
	variable i
	colorList="keyword:0000ff;comment:ff0000;string:009c00;"
	colorList+="operation:007575;function:c34e00;pound:cc00a3"
	for(i=0;i<itemsinlist(colorList);i+=1)
		string color,colorType=stringfromlist(0,stringfromlist(i,colorList),":")
		if(userColors)
			Execute /Q "SetIgorOption colorize, "+color+"Color=?"
			NVAR V_red, V_green, V_blue
			sprintf color, "%.2x%.2x%.2x", RedBD(V_red), RedBD(V_green), RedBD(V_blue)
			colorList=ReplaceStringByKey(colorType,colorList,color)
		endif
	endfor
	
	// Generate key-value pairs for word types and the words belonging to each type.  
	wordList="flow:;types:;functions:;operations:;pragmas:"
	for(i=0;i<itemsinlist(wordList);i+=1)
		string words,wordType=stringfromlist(0,stringfromlist(i,wordList),":")
		strswitch(wordType)
			case "flow":
				words=FlowControlList()
				break
			case "types":
				words=TypesList()
				break
			case "functions":
				words=FunctionList("*", ";", "KIND:1")
				break
			case "operations":
				words=OperationList("*", ";", "internal")
				break
			case "pragmas":
				words=PragmasList()
				break
		endswitch
		wordList=ReplaceStringByKey(wordType,wordList,ReplaceString(";",words," ")) // The words will be separated by spaces instead of semi-colons, as this works better with the key-value format.  
	endfor
	
	return 0
End

static Function SetXMLValues(type,fileID,words,colors)
	string type // Type of XML file, i.e. the text editor with which we are dealing.  
	variable fileID
	string words
	string colors
	
	XMLelemlist(fileID) // Get a text wave of XML elements.  
	wave /t elements=w_elementlist
	
	strswitch(type)
		case "NotepadPlus": // The XML file is used by Notepad++.  
			// Find the part of the XML file that contains information about Igor syntax coloring, so we don't overwrite any other custom syntax coloring.  
			duplicate /free /r=[][2] w_elementlist,name
			findvalue /TEXT="igor" /TXOP=0 name
			variable nodeNum=v_value
			string base=elements[nodeNum][0]
			
			XMLsetAttr(fileID,base,"","ext","ipf") // Make this apply automatically to all .ipf files opened in the text editor.  
			XMLsetAttr(fileID,base+"/Settings/Global","","caseIgnored","yes") // Case insensitive
			XMLsetAttr(fileID,base+"/Settings/Global","","escapeChar","\\") // Escape character
			XMLsetNodeStr(fileID,base+"/KeywordLists/Keywords[1]","","\"00\"00") // Delimiters
			XMLsetNodeStr(fileID,base+"/KeywordLists/Keywords[4]","","! \" # $ & ( : ? , [ ] { } + =") // Operators
			XMLsetNodeStr(fileID,base+"/KeywordLists/Keywords[5]",""," 1 2 0//") // Comment
			XMLsetNodeStr(fileID,base+"/KeywordLists/Keywords[6]","",stringbykey("flow",words)+stringbykey("types",words)) // Words 1
			XMLsetNodeStr(fileID,base+"/KeywordLists/Keywords[7]","",stringbykey("functions",words)) // Words 2
			XMLsetNodeStr(fileID,base+"/KeywordLists/Keywords[8]","",stringbykey("operations",words)) // Words 3
			XMLsetNodeStr(fileID,base+"/KeywordLists/Keywords[9]","",stringbykey("pragmas",words)) // Words 4
			XMLsetAttr(fileID,base+"/Styles/WordsStyle[4]","","fgColor",stringbykey("keyword",colors)) // Keyword 1
			XMLsetAttr(fileID,base+"/Styles/WordsStyle[5]","","fgColor",stringbykey("function",colors)) // Keyword 2
			XMLsetAttr(fileID,base+"/Styles/WordsStyle[6]","","fgColor",stringbykey("operation",colors)) // Keyword 3
			XMLsetAttr(fileID,base+"/Styles/WordsStyle[7]","","fgColor",stringbykey("pound",colors)) // Keyword 4
			XMLsetAttr(fileID,base+"/Styles/WordsStyle[9]","","fgColor",stringbykey("comment",colors)) // Comment Line
			XMLsetAttr(fileID,base+"/Styles/WordsStyle[12]","","fgColor",stringbykey("string",colors)) // Deliminer1
			break
	endswitch
	
	killwaves /z w_elementlist
	return 0
End

static StrConstant kBlack = "#000000"
static Function RedBD(num)
	// Reduces the bit depth of a number from 16 bit to 8 bit
	Variable num
	num /= 2^16
	num *= 2^8
	num = floor(num)
	return num
End

static Function/S FlowControlList()
	String list = ""
	// flow control	
	list += "if;else;elseif;endif;for;endfor;"
	list += "strswitch;switch;case;endswitch;"
	list += "do;while;try;catch;endtry;break;continue;return;"
	return list
End

static Function/S TypesList()
	String list = ""
	// data types
	list += "variable;string;constant;strconstant;"
	// reference types
	list += "NVAR;SVAR;WAVE;STRUCT;"
	// function/procedure/macro types
	list += "function;end;static;macro;window;graph;Structure;EndStructure;EndMacro;"
	list += "FuncFit;Proc;Picture;Menu;SubMenu;"
	// Commands
	list += "Prompt;DoPrompt;"
	return list
End

static Function/S PragmasList()
	String list = "#include;#pragma;"
	return list
End